/*
 * Draw a UCC Calendar Clock on a canvas initialised to today's date and time.
 *
 * Animate the calendar if required, set any UCC or Gregorian date.
 *
 * by Prajna Pranab, Leo 13517
 *
 * Developed for Web app and Android/iOS hybrid app.
 */

(function(){
  const VERSION = '1.2.2a';
  const DEG_PER_YR = 360 / 24000;   // fraction of a degree per year discounting precession
  const CIRC = 2 * Math.PI;
  const MONTHS = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
  var currentDate = new UCCDate();  // so we can start and stop the animation from a particular date
  var animSpeed = 1000;             // 1000ms = 1 second
  var callBack = 0;                 // animation callback
  var saveCount = 0;                // track context saves and restores
  var heading = 0;                  // current rotation of the clock face.
  
  // update options from local storage if they've been saved
  var options = (localStorage.options) ? JSON.parse(localStorage.options) : currOptions();
  // reconfigure the interface to match the options
  setOptions();
  updateTitles();
  
  // save width of controls window for later comparison
  var ctlWidth = Math.round(Id('controls').getBoundingClientRect().width);
  
  // calculate a scaling factor from the window size
  var minAspect = (Math.min(document.documentElement.clientWidth, document.documentElement.clientHeight));
  var zoomFactor = minAspect / 1100; // drawing was designed at 1100px
  var zoom = floor(zoomFactor * 100) + '%';
  // html zoom does a better job than canvas scale
  document.body.style.zoom = zoom;
  // FF doesn't support html zoom
  document.body.style.MozTransform = 'scale(' + zoomFactor + ')';
  document.body.style.MozTransformOrigin = "center top";
  
  // get the drawing context
  var ctx = Id('clockCanvas').getContext("2d");
  
  // get canvas size and center
  var canvas = {
    width: ctx.canvas.width,
    height: ctx.canvas.height,
    cx: ctx.canvas.width / 2,
    cy: ctx.canvas.height / 2
  }
    
  // load the background images, asp
  var zodiac = new Image();
  zodiac.onload = imagesReady;
  zodiac.src = 'asset/images/zodiac.png';
  var earth = new Image();
  earth.onload = imagesReady;
  earth.src = 'asset/images/Earth-01.png';
  
  // set default styles
  var defaults = {
    shadowOffsetX: 2,
    shadowOffsetY: 2,
    shadowBlur: 2,
    shadowColor: "rgba(100, 50, 50, 1)",
    fontFamily: "sans-serif",
    fontSize: 14,
    fillStyle: 'black',
    strokeStyle: 'black',
    lineWidth: 1,
    lineDash: [1, 0],
    lineHeight: 22,
    textAlign: 'center',
    stroke: true,
    fill: true
  };
  
  // set up touch/mouse event handlers
  eventHandlers();
  
  // start drawing once images have loaded
  function imagesReady(evt) {
    if (earth.complete && zodiac.complete) {
      Id('progress').style.display = 'none';
      drawCalendar();
    }
  }
  
  function drawCalendar(today) {
    
    // update the options and checkbox titles
    options=currOptions();
    updateTitles();
  
    // set the date
    today = today || currentDate;     // by default carry on from where we were
    today.names = options.language;   // set the language for deekday/triad names
    currentDate = today;  // so we can animate from this date
    
    // tidy up any unrestored saves to reset to original canvas state
    if (saveCount) while (saveCount--) ctx.restore();
    
    // save the context
    save();
    
    // clear any translates
    // ctx.setTransform(1, 0, 0, 1, 0, 0);
    
    // clear the canvas
    ctx.clearRect(0, 0, canvas.width, canvas.height);
    
    // draw the upper date info table
    var dateStrs = [today.full, today.long, today.medium, today.short, today.sortable];
    boxText({
      padding: 10,
      pos: 'tr',
      textAlign: 'right',
      fillStyle: "rgba(256, 100, 100, 0.8)",
      strokeStyle: 'gray',
      strs: [today.yuga,
             today.zodiac,
             'Moon: ' + today.moon,
             dateStrs[options.dateFmt],
             today.jDate.getDate() + ' ' + MONTHS[today.jDate.getMonth()] + ' ' + today.jDate.getFullYear()]
    });
    
    // draw the lower date info table
    boxText({
      padding: 10,
      pos: 'br',
      textAlign: 'right',
      fillStyle: "rgba(256, 100, 100, 0.8)",
      strokeStyle: 'gray',
      strs: ['Decan No: ' + today.deekNumber,
             'Leap Year: ' + (today.leapYear ? 'yes' : 'no'),
             'Day of Year No: ' + today.doy,
             'Gt Year Day No: ' + commafy(today.days)]
    });
    
    // draw the data table
    boxText({
      padding: 10,
      pos: 'tl',
      textAlign: 'left',
      fillStyle: "rgba(256, 100, 200, 0.5)",
      strokeStyle: 'gray',
      strs: ['zoom: ' + zoom,
             '<Swipe right here for settings>']
    });
    
    // draw the data table
    boxText({
      padding: 10,
      pos: 'bl',
      textAlign: 'left',
      fillStyle: "rgba(256, 100, 100, 0.8)",
      strokeStyle: 'gray',
      strs: ['Lib Ver: ' + today.version,
             'Software Ver: ' + VERSION,
             'Copyright © The Troubadour Community Trust']
    });
    
    // move the origin to the center of the canvas
    ctx.translate(canvas.cx, canvas.cy);
    
    // draw and rotate text from its center
    ctx.textBaseline = 'middle';
    
    // rotate to year 12000
    setHeading(12000);
    
    // draw the Great Year disc
    doGreatYear({
      radius: 500,
      marker: {
          radius: 500,
          length: 120,
          lineWidth: 2,
          shadowColor: "rgba(256, 100, 100, 1)"
      },
      caption: {
          radius: 440,
          shadowColor: "rgba(256, 100, 100, 0.8)"
      },
      golden: {
          length: 4800,
          descending: {start: 0},
          ascending: {start: 19200},
          grad: [0, 0, 0, 0, 0, 500],
          stops: [[0.75,'rgba(255, 200, 0, 0)'], [0.85, 'rgba(255, 200, 0, 1)'], [1, 'rgba(255, 200, 0, 0)']]
      },
      silver: {
          length: 3600,
          descending: {start: 4800},
          ascending: {start: 15600},
          grad: [0, 0, 0, 0, 0, 500],
          stops: [[0.75,'rgba(255, 255, 255, 0)'],[0.85, 'rgba(255, 255, 255, 1)'], [1, 'rgba(255, 255, 255, 0)']]
      },
      bronze: {
          length: 2400,
          descending: {start: 8400},
          ascending: {start: 13200},
          grad: [0, 0, 0, 0, 0, 500],
          stops: [[0.75,'rgba(200, 100, 0, 0)'],[0.85, 'rgba(200, 100, 0, 1)'], [1, 'rgba(200, 100, 0, 0)']]
      },
      iron: {
          length: 1200,
          descending: {start: 10800},
          ascending: {start: 12000},
          grad: [0, 0, 0, 0, 0, 500],
          stops: [[0.75,'rgba(120, 100, 100, 0)'],[0.85, 'rgba(120, 100, 100, 1)'], [1, 'rgba(120, 100, 100, 0)']]
      }
    });
    
    if (options.sandhis) doSandhis({
        radius: 500,
        length: 120,
        lineDash: [3, 6],
        shadowColor: "rgba(256, 100, 100, 1)",
        years: [400, 4400, 5100, 8100, 8600, 10600, 10900, 11900, 12100, 13100, 13400, 15400, 15900, 18900, 19600, 23600]
    });
    
    // reset rotation to draw constellations etc.
    setHeading(12000);
    
    // draw constellation background
    doConstellations({
        radius: 400,
        length: 30,
        lineWidth: 2,
        fillStyle: 'white',
        strokeStyle: 'white',
        shadowColor: "rgba(256, 100, 100, 1)",
        years: [0, 1833, 3733, 5699, 7732, 9832, 12000, 14168, 16268, 18301, 20267, 22167]
    });
    
  // do the day disc 1st so the calendar overlaps it
    dayDisc({
      radius: 288,
      size: 60,
      grads: {
          grad: [0, 0, 0, 0, 0, 60],
          stops: [[0,'orange'],[0.3,'orange'],[0.7, 'yellow'], [0.95, '#885138'], [1, '#885138']]
      },
      disc: {
        shadowColor: "rgba(256, 100, 100, 1)",
        fillStyle: 'yellow',
        strokeStyle: 'cyan'
      },
      cog: {
        strokeStyle: 'white'
      },
      ring: {
        strokeStyle: 'gray'
      },
      text: {
        strokeStyle: 'cyan'
      }
    });
   
    // define and draw discs
    // blue disc
    doDisc({
      radius: 235,
      grads: {
          grad: [0, 0, 175, 0, 0, 240],
          stops: [[0,'#98CCFE'],[0.5, '#99fffd'], [1, '#999bff']]
        }
    });
    
    // red disc
    doDisc({
      radius: 175,
      grads: {
          grad: [0, 0, 142, 0, 0, 180],
          stops: [[0,'#ebb73d'],[0.2, '#eb3d71'], [0.5, '#eb603c'], [1, '#ebb73d']]
        }
    });
    
    // white disc
    doDisc({
      radius: 142,
      grads: {
          grad: [0, 0, 120, 0, 0, 135],
          stops: [[0,'#e0e0e0'],[0.5, '#ffffff'], [1, '#f0f0f0']]
        }
    });
    
    // green disc
    doDisc({
      radius: 115,
      grads: {
          grad: [0, 0, 78, 0, 0, 112],
          stops: [[0,'#bab45b'],[0.2, '#91ba5b'], [0.9, '#62ba5b'], [1, '#bab45b']]
        }
    });
    
    // Great Year to-Date arrow
    if (options.gysf) doGtYrMarker({
      radius: 420,
      length: 5,
      shadowColor: "rgba(100, 50, 50, 0)",
      grads: {
          grad: [0, 0, 395, 0, 0, 420],
          stops: [[0,'rgba(100, 250, 100, 0)'],
                  [0.1, 'rgba(100, 250, 100, 1)'],
                  [0.2, 'rgba(100, 250, 100, 0.75)'],
                  [0.3, 'rgba(100, 250, 100, 0.25)'],
                  [0.4, 'rgba(100, 250, 100, 0.01)'],
                  [0.6, 'rgba(100, 250, 100, 0)']]
        }
    });
    
    // this year marker
    yearMarker({
      radius: 500,
      length: 170,
      shadowColor: "rgba(256, 100, 100, 1)",
      fontSize: 24,
      fillStyle: 'white',
      strokeStyle: 'white',
      grad: [0, 0, 50, 0, 0, 500],
      stops: [[0.75,'rgba(100, 250, 100, 0)'],[0.85, 'rgba(100, 250, 100, 1)'], [1, 'rgba(100, 250, 100, 0)']]
    });
    
    // spinning earth
    doEarth({
      radius: 80
    });
    
    // sun shadow on earth
    earthShadow({
      radius: 70,
      grads: {
          grad: [0, 0, 0, 0, 0, 80],
          stops: [[0,'rgba(50, 50, 50, 0.0)'], [0.2, 'rgba(50, 50, 50, 0.8)'], [0.5, 'rgba(50, 50, 50, 0.8)'], [1, 'rgba(0, 0, 0, 1)']]
        }
    });
          
    // draw text inside the earth disc
    doSunText({
      shadowColor: "rgba(0, 0, 200, 0.8)",
      fontSize: 32,
      fillStyle: 'white',
      strokeStyle: 'white'
    });
    
    // draw darts
    doDarts({
       radius: 34,
       length: 28,
       width: 10,
       normal: {
         lineWidth: 2,
         strokeStyle: 'yellow'
       },
       special: {
         lineWidth: 3,
         fillStyle: 'yellow'
       }
    });
    
    // draw triad rays
    doTriadRays({
      radius: 235,
      length: 165
    });
    
    // draw decan rays
    doDeekRays({
      radius: 175,
      length: 33
    });
    
    // draw cardinal discs and rays
    doCardRays({
      radius: 195,
      length: 123,
      line1: {
        lineWidth: 6,
        fillStyle: 'yellow',
        strokeStyle: 'yellow'
      },
      line2: {
        lineWidth: 2
      },
      normal: {
        radius: 20,
        pos: 215,
        grads: {
          grad: [0, 0, 0, 2, -4, 20],
          stops: [[0,'#98CCFE'],[0.5, '#99fffd'], [1, '#999bff']]
        }
      },
      special: {
        radius: 20,
        lineWidth: 2,
        grads: {
          grad: [0, 0, 0, 2, -4, 20],
          stops: [[0,'#FFFF00'],[0.5, '#eeee00'], [1, '#ffff99']]
        }
      },
      style: {
        shadowOffsetX: 3,
        shadowOffsetY: 3,
        shadowBlur: 4,
        shadowColor: "rgba(0, 0, 200, 1)"
      },
      strs: ['2', '93', '184', '275']
    });
    
    // draw Triad numbers
    doTriadNumbers({
      radius: 90,
      normal: {
        shadowOffsetX: 3,
        shadowOffsetY: 3,
        shadowBlur: 4,
        shadowColor: "rgba(0, 0, 200, 1)",
        strokeStyle: 'yellow'
      },
      special: {
        shadowOffsetX: 3,
        shadowOffsetY: 3,
        shadowBlur: 4,
        shadowColor: "rgba(255, 255, 0, 1)",
        fontSize: 26,
        fillStyle: 'cyan',
      },
      spotLight: {
        radius: 20,
        shadowColor: 'rgba(255, 255, 0, 0)',
        grads: {
          grad: [0, 0, 0, 0, 0, 20],
          stops: [[0,'rgba(255, 255, 255, 0.8'],[0.5, 'rgba(255, 255, 0, 0.5'], [1, 'rgba(255, 255, 0, 0)']]
        }
      }
    });
    
    // draw Decan numbers
    doDeekNumbers({
      radius: 160,
      normal: {
        shadowOffsetX: 3,
        shadowOffsetY: 3,
        shadowBlur: 4,
        shadowColor: "rgba(0, 0, 200, 1)",
        fillStyle: 'white',
        strokeStyle: 'yellow'
      },
      special: {
        shadowOffsetX: 3,
        shadowOffsetY: 3,
        shadowBlur: 4,
        shadowColor: "rgba(255, 255, 0, 1)",
        fontSize: 18,
        fillStyle: 'cyan'
      },
      spotLight: {
        radius: 20,
        shadowColor: 'rgba(255, 255, 0, 0)',
        grads: {
          grad: [0, 0, 0, 0, 0, 20],
          stops: [[0,'rgba(255, 255, 255, 0.8'],[0.5, 'rgba(255, 255, 0, 0.5'], [1, 'rgba(255, 255, 0, 0)']]
        }
      }
    });
    
    // draw Triad symbols
    doTriadSymbols({
      radius: 205,
      normal: {
        shadowOffsetX: 3,
        shadowOffsetY: 3,
        shadowBlur: 4,
        shadowColor: "rgba(0, 0, 200, 1)",
        fontSize: 24,
        fillStyle: 'yellow',
        strokeStyle: 'yellow'
      },
      special: {
        shadowOffsetX: 3,
        shadowOffsetY: 3,
        shadowBlur: 8,
        shadowColor: "rgba(255, 255, 0, 1)",
        fontSize: 32,
        fillStyle: 'cyan',
        strokeStyle: '#333'
      },
      spotLight: {
        radius: 20,
        shadowColor: 'rgba(255, 255, 0, 0)',
        grads: {
          grad: [0, 0, 0, 0, 0, 20],
          stops: [[0,'rgba(255, 255, 255, 0.8'],[0.5, 'rgba(255, 255, 0, 0.5'], [1, 'rgba(255, 255, 0, 0)']]
        }
      }
    });
    
    if (options.moon) doMoon({
      radius: 130,
      size: 20,
      grads: {
          grad: [0, 0, 0, 2, 2, 28],
          stops: [[0,'#909000'],[0.9, '#ffffff'], [1, '#f0f0f0']]
        },
      shadow: {
        radius: 20,
        grads: {
            grad: [0, 0, 0, 0, 0, 80],
            stops: [[0,'rgba(50, 50, 50, 0.0)'], [0.2, 'rgba(50, 50, 50, 0.8)'], [0.5, 'rgba(50, 50, 50, 0.8)'], [1, 'rgba(0, 0, 0, 1)']]
          }
      }
    });
    
    restore();
    
    //**************************** Helper Functions ************************
    
    // keep track of how many saves we've done
    function save() {
    	ctx.save();
    	saveCount++;
    }
    
    // keep track of how many restores we've done
    function restore() {
    	ctx.restore();
      saveCount--;
    }
    
    // return degrees for a given number of radians (anti-clockwise)
    function rad2deg(rads) {
      return -(rads  * 180 / Math.PI);
    }
    
    // return radians for a given number of degrees (anti-clockwise)
    function deg2rad(deg) {
      return -(deg  * Math.PI / 180.0);
    }
    
    // return radian angle for a given year
    function yr2rad(year) {
    	return deg2rad(DEG_PER_YR * year);
    }
    
    // return radian angle for a given year accounting for vrp if set
    function yr2radP(year) {
      return (options.vrp ? yearPos(year) : yr2rad(year));
    }
    
    // convert a day number into radians
    function day2rad(day) {
    	return deg2rad(today.doy - today.intercals);
    }
    
    // return an orbital position, in radians, for a given year, orbital period and elliptical eccentricity
    // Thanks to http://nbodyphysics.com/blog/2016/05/29/planetary-orbits-in-javascript/
    // Here be the rocket science. You have been warned!
    function orbit(year, period, eccentricity) {
    	var u = M = 2.0 * Math.PI * year/period; // M is the position of a body moving in a uniform circle at a constant rate
    	var n = new_u = 0;
    	var limit = 10;
  	  // calculate the difference between a circular and elliptical orbit for this point in the orbit
    	while (n++<limit) {
    	  new_u = u + (M - (u - eccentricity * Math.sin(u)))/(1 - eccentricity * Math.cos(u));
    	  // bail out if required precision is reached
    		if (Math.abs(new_u - u) < 1E-6)
    			break;
    		u = new_u;
    	}
    	return -u;
    }
    
    // return a precession-adjusted great year orbital position
    function yearPos(year) {
    	return orbit(year, 24000, 0.08);
    }
    
    // return a precession-adjusted lunar orbital position
    function lunaPos() {
      var periods = {
        synodic: 29.530588853,
        sidereal: 27.321662,
        tropical: 27.321582,
        anomilistic: 27.554550,
        draconic: 27.212221
      }
      // get the difference in days between this date and the full moon eclipse date of 11 Aug 1999
      var fullMoon = new UCCDate(Date.UTC(1999, 7, 11));
      var days = Math.abs(today.days - fullMoon.days);
      // get orbit adjusted to line up with full moon position on 11 Aug 1999
    	return orbit(days, periods.sidereal, 0.0549) + deg2rad(140);
    }
    
    // set and save a heading (direction)
    function setHeading(year) {
      heading = (options.vrp ? yearPos(year) : yr2rad(year));
    	ctx.rotate(heading);
    }
    
    // fill or stroke text horizontally at a given radius
    function drawText(str, settings) {
      save();
    	ctx.translate(settings.radius, 0);
    	// rotate to write text oriented horizontally when rotated back to 0
    	ctx.rotate(deg2rad(180) - heading);
    	setStyles(settings);
    	if (settings.fill) ctx.fillText(str, 0, 0);
    	if (settings.stroke) ctx.strokeText(str, 0, 0);
    	restore();
    }
    
    // set any unspecified styles to defaults and configure ctx styles
    function setStyles(settings) {
      function getFont(settings) {
        return settings.fontSize + 'pt ' + settings.fontFamily;
      }
      settings.lineHeight = settings.lineHeight || defaults.lineHeight;
      settings.fill = settings.fill || defaults.fill;
      settings.stroke = settings.stroke || defaults.stroke;
      settings.fontFamily = settings.fontFamily || defaults.fontFamily;
      settings.fontSize = settings.fontSize || defaults.fontSize;
      ctx.shadowOffsetX = settings.shadowOffsetX || defaults.shadowOffsetX;
      ctx.shadowOffsetY = settings.shadowOffsetY || defaults.shadowOffsetY;
      ctx.shadowBlur = settings.shadowBlur || defaults.shadowBlur;
      ctx.shadowColor = settings.shadowColor || defaults.shadowColor;
      ctx.font = getFont(settings);
      ctx.fillStyle = settings.fillStyle || defaults.fillStyle;
      ctx.strokeStyle = settings.strokeStyle || defaults.strokeStyle;
      ctx.lineWidth = settings.lineWidth || defaults.lineWidth;
      ctx.setLineDash(settings.lineDash || defaults.lineDash);
      ctx.textAlign = settings.textAlign || defaults.textAlign;
    }
    
    // create gradient from a settings data structure
    function grads(settings) {
      // scale the gradient radii
      var g = settings.grad.slice();
      var gradient = ctx.createRadialGradient(g[0], g[1], g[2], g[3], g[4], g[5]);
      for (var n=0; n<settings.stops.length; n++)
      	gradient.addColorStop(settings.stops[n][0], settings.stops[n][1]);
     return gradient;
    }
    
    // draw a line on the clock-face at a specified rotation
    function marker(year, settings) {
      save();
      setStyles(settings);
      setHeading(year);
      ctx.beginPath();
      ctx.moveTo(settings.radius - settings.length, 0);
      ctx.lineTo(settings.radius, 0);
      ctx.stroke();
      restore();
    }
      
    function doDisc(settings) {
      save();
      setStyles(settings);
      ctx.fillStyle = grads(settings.grads);
      ctx.beginPath();
      ctx.arc(0, 0, settings.radius, 0, CIRC, true);
      if (settings.fill) ctx.fill();
      if (settings.stroke) ctx.stroke();
      restore();
    }
    
    //******************************* Drawing Functions **************************
    
    function boxText(settings, n) {
      function getX(settings) {
      	return (settings.pos == 'tl' || settings.pos == 'bl') ?
      		settings.padding :
      	  canvas.width - settings.padding;
      }
      function getY(settings, n) {
      	return (settings.pos == 'tl' || settings.pos == 'tr') ?
      		settings.lineHeight * (n+1) :
      	  canvas.height - (settings.lineHeight * (settings.strs.length - n - 1)) - settings.padding;
      }
    	save();
    	setStyles(settings);
    	for (var n=0; n<settings.strs.length; n++) {
    	  var x = getX(settings);
    	  var y = getY(settings, n);
    		if (settings.fill) ctx.fillText(settings.strs[n], x, y);
  		  if (settings.stroke) ctx.strokeText(settings.strs[n], x, y);
    	}
    	restore();
    }
    
    function doGreatYear(settings) {
      function arc(radius, settings) {
        save();
        // draw descending arc
        ctx.beginPath();
        ctx.arc(0, 0, radius, yr2radP(settings.descending.start), yr2radP(settings.descending.start + settings.length), true);
        ctx.arc(0, 0, 0, yr2radP(settings.descending.start), yr2radP(settings.descending.start + settings.length), true);
        ctx.fillStyle = grads(settings);
        ctx.fill();
        // draw ascending arc
        ctx.beginPath();
        ctx.arc(0, 0, radius, yr2radP(settings.ascending.start), yr2radP(settings.ascending.start + settings.length), true);
        ctx.arc(0, 0, 0, yr2radP(settings.ascending.start), yr2radP(settings.ascending.start + settings.length), true);
        ctx.fill();
        restore();
      }
      function yearCaption(year, settings) {
        save();
        setStyles(settings);
        // move year captions off the horizontal line
        if (year==0)
          setHeading(year + 120);
        else if (year==24000)
        	setHeading(year - 80);
        else if (year==12000)
        	setHeading(year + 80);
        else
        	setHeading(year);
        drawText(year, settings)
        restore();
      }
      save();
    	// draw the golden age arc
    	arc(settings.radius, settings.golden);
      // silver age
    	arc(settings.radius, settings.silver);
      // bronze age
    	arc(settings.radius, settings.bronze);
      // iron age arc
    	arc(settings.radius, settings.iron);
      // year 0 marker
      marker(0, settings.marker);
      // year 12,000 marker
      marker(12000, settings.marker);
      // write year numbers
      yearCaption(24000, settings.caption);
      yearCaption(settings.golden.descending.start, settings.caption);
      yearCaption(settings.golden.ascending.start, settings.caption);
      yearCaption(settings.silver.descending.start, settings.caption);
      yearCaption(settings.silver.ascending.start, settings.caption);
      yearCaption(settings.bronze.descending.start, settings.caption);
      yearCaption(settings.bronze.ascending.start, settings.caption);
      yearCaption(settings.iron.descending.start, settings.caption);
      yearCaption(settings.iron.ascending.start, settings.caption);
    	// restore original rotation */
    	restore();
    }
    
    function doSandhis(settings) {
      save();
      for (var n=0; n<settings.years.length; n++) {
      	marker(settings.years[n], settings);
      }
      restore();
    }
    
    function doConstellations(settings) {
      function constMarkers(settings) {
      	save();
      	for (var n=0; n<settings.years.length; n++)
      		marker(settings.years[n], settings);
      	restore();
      }
    
    	save();
    	var r = settings.radius;
      // clip around the zodiac constellation background
      ctx.beginPath();
      ctx.arc(0, 0, r, 0, CIRC, true);
      ctx.clip();
      
      // rotate so 1 Aries points at today
      setHeading(today.year);
      
      // draw the background
      ctx.drawImage(zodiac, -r, -r, 2 * r, 2 * r);
      
      if (options.constMarkers) constMarkers(settings);
              
    	restore();
    	
    }
    
    function doGtYrMarker(settings) {
      save();
      setHeading(12000);
      setStyles(settings);
      ctx.beginPath();
      ctx.arc(0, 0, settings.radius, yr2rad(30), yr2radP(today.year- 30), true);
      ctx.arc(0, 0, settings.radius - settings.length, yr2rad(30), yr2radP(today.year - 30), true);
      ctx.fillStyle = grads(settings.grads);
      ctx.fill();
      restore();
    }
    
    function yearMarker(settings) {
    	save();
    	// rotate year
    	setHeading(12000)
    	setHeading(today.year);
      ctx.beginPath();
      // draw marker
      setStyles(settings);
      ctx.arc(0, 0, settings.radius, deg2rad(5), deg2rad(-5));
      ctx.arc(0, 0, settings.radius - settings.length, deg2rad(5), deg2rad(-5));
      ctx.fillStyle = grads(settings);
      ctx.fill();
      // write the year number
      drawText(today.year, settings);
    	restore();
    }
    
    function doEarth(settings) {
      // draw the earth
      var r = settings.radius;
      ctx.drawImage(earth, -r, -r, 2 * r, 2 * r);
    }
    
    function earthShadow(settings) {
    	save();
    	ctx.fillStyle = grads(settings.grads)
    	ctx.rotate(day2rad(today.day - today.intercals));
    	ctx.beginPath();
    	ctx.arc(0, 0, settings.radius, deg2rad(-90), deg2rad(90));
    	ctx.arc(0, -settings.radius, 2, deg2rad(-90), deg2rad(90));
    	ctx.fill();
    	restore();
    }
    
    function dayDisc(settings) {
      function cog(settings) {
        save(); // before rotating again
        var rot = today.doy - today.intercals;
        // synchronize the cog with the clock-face
        ctx.rotate(deg2rad(rot * 4));
        ctx.beginPath();
        setStyles(settings.cog);
        ctx.fillStyle = grads(settings.grads);
        for (var n=1; n<37; n++) {
          var deg = n * 10;
          if (n % 2) {
            ctx.arc(0, 0, settings.size, deg2rad(deg), deg2rad(deg + 7), true);
          } else {
            ctx.arc(0, 0, settings.size - 5, deg2rad(deg), deg2rad(deg + 7), true);
          }
        }
        ctx.stroke();
        ctx.fill();
        // ring inside gear teeth
        ctx.beginPath();
        ctx.strokeStyle = settings.ring.strokeStyle;
        ctx.arc(0, 0, settings.size - 10, 0, CIRC, true);
        ctx.stroke();
        restore();
      }
      save();
      // rotate the clock to where the day is
      var rotation = day2rad(today.doy - today.intercals);
      ctx.rotate(rotation);
      // move into position
      ctx.translate(settings.radius, 0);
      setStyles(settings.disc);
      // draw the cog
      cog(settings);
      // rotate before drawing text
    	ctx.rotate(0 - rotation);
      setStyles(settings.text);
      var str = today.day;
      var lh = settings.text.lineHeight;
      var top = -lh + 10;
    	ctx.fillText(str, 0, top - lh);
    	// draw deekday symbol or not
    	if (today.intercal || !options.deekSymb) {
        str = (today.intercal) ? today.intercal : today.deekDay;
      	ctx.fillText(str, 0, top + (lh / 2));
    	} else {
        str = today.deekSymbol;
      	ctx.fillText(str, 0, top);
        str = today.deekDay;
      	ctx.fillText(str, 0, top + lh);
    	}
    	str = today.doy;
    	ctx.fillText(str, 0, top + (2 * lh));
    	restore();
    }
    
    function spotLight(settings) {
    	save();
      setStyles(settings);
      ctx.fillStyle = grads(settings.grads);
      ctx.beginPath();
      ctx.arc(0, 0, settings.radius, 0, CIRC, true);
      ctx.fill();
    	restore();
    }
    
    function doTriadSymbols(settings) {
      save();
      // rotate back a little to center the numbers
    	ctx.rotate(deg2rad(15));
      for (n=1; n<13; n++) {
      	// move out to text position
        ctx.translate(settings.radius, 0);
        // style the text
        setStyles(settings.normal);
        if (today.triad == n) {
          spotLight(settings.spotLight);
          setStyles(settings.special);
        }
        // rotate to draw text in correct orientation
      	ctx.rotate(deg2rad((n * -30) + 15));
      	ctx.fillText(today.TSYMBOLS[n -1], 0, 2);
      	ctx.strokeText(today.TSYMBOLS[n -1], 0, 2);
      	// then move back before rotating the clock
      	ctx.rotate(deg2rad((n * 30) - 15));
        ctx.translate(-settings.radius, 0);
      	ctx.rotate(deg2rad(30));
      }
      restore();
    }
    
    function doDeekNumbers(settings) {
      save();
      // rotate back a little to center the numbers
    	ctx.rotate(deg2rad(5));
      for (n=1; n<37; n++) {
      	// move out to text position
        ctx.translate(settings.radius, 0);
        // style the text
        setStyles(settings.normal);
        if (today.deekNumber == n) {
          spotLight(settings.spotLight);
          setStyles(settings.special);
        }
        // rotate before drawing text
      	ctx.rotate(deg2rad((n * -10) + 5));
      	ctx.fillText(n, 0, 2);
       	if (today.deekNumber == n) ctx.strokeText(n, 0, 2);
       	// then move back before rotating the clock
      	ctx.rotate(deg2rad((n * 10) - 5));
        ctx.translate(-settings.radius, 0);
      	ctx.rotate(deg2rad(10));
      }
      restore();
    }
    
    function doTriadNumbers(settings) {
      save();
      // rotate back a little to center the numbers
    	ctx.rotate(deg2rad(15));
      for (n=1; n<13; n++) {
        // move out to where the number will be
        ctx.translate(settings.radius, 0);
        // style the text
        setStyles(settings.normal);
        if (today.triad == n) {
          spotLight(settings.spotLight);
          setStyles(settings.special);
        }
      	// rotate before drawing text
      	ctx.rotate(deg2rad((n * -30) + 15));
      	ctx.fillText(n, 0, 2);
      	if (today.triad == n) ctx.strokeText(n, 0, 2);
      	// then move back before rotating the clock
      	ctx.rotate(deg2rad((n * 30) - 15));
        ctx.translate(-settings.radius, 0);
      	ctx.rotate(deg2rad(30));
      }
      restore();
    }
    
    function doTriadRays(settings) {
      function line(settings) {
        ctx.beginPath();
        setStyles(settings);
        ctx.moveTo(settings.radius - settings.length, 0);
        ctx.lineTo(settings.radius, 0);
        ctx.stroke();
      }
      
      save();
      for (var n=0; n<12; n++) {
      	line(settings);
      	ctx.rotate(deg2rad(30));
      }
      restore();
    }
    
    function doCardRays(settings) {
      function line() {
        save();
        setStyles(settings.line1);
        ctx.beginPath();
        ctx.moveTo(settings.radius - settings.length, 0);
        ctx.lineTo(settings.radius, 0);
        ctx.stroke();
        setStyles(settings.line2);
        ctx.beginPath();
        ctx.moveTo(settings.radius - settings.length, 0);
        ctx.lineTo(settings.radius, 0);
        ctx.stroke();
        restore();
      }
      
      save();
      for (var n=0; n<4; n++) {
        setStyles(settings.style);
      	line();
      	// move out to do the disc and text
        ctx.translate(settings.normal.pos, 0);
        if (today.doy == Number(settings.strs[n]))
       	  doDisc(settings.special);
       	else
       	  doDisc(settings.normal);
      	// rotate to draw text
      	ctx.rotate(deg2rad(n * -90));
      	ctx.fillText(settings.strs[n], 0, 0);
      	ctx.rotate(deg2rad(n * 90));
        ctx.translate(-settings.normal.pos, 0);
      	ctx.rotate(deg2rad(90));
      }
      restore();
    }
    
    function doDeekRays(settings) {
      function line() {
        ctx.beginPath();
        ctx.moveTo(settings.radius - settings.length, 0);
        ctx.lineTo(settings.radius, 0);
        ctx.stroke();
      }
      
      save();
      for (var n=0; n<36; n++) {
      	line();
      	ctx.rotate(deg2rad(10));
      }
      restore();
    }
    
    function doSunText(settings) {
      save();
      setStyles(settings);
      var str = (today.doy) ? '1' : '0';
      if (settings.fill) ctx.fillText(str, 0, 5);
      if (settings.stroke) ctx.strokeText(str, 0, 5);
      restore();
    }
    
    function doDarts(settings) {
      function dart(special) {
        save();
        setStyles(settings.normal);
        ctx.beginPath();
        ctx.moveTo(settings.radius, -settings.width);
        ctx.lineTo(settings.radius, settings.width);
        ctx.lineTo(settings.radius + settings.length, 0);
        ctx.lineTo(settings.radius, -settings.width);
        if (special) {
          setStyles(settings.special);
          ctx.fill();
        }
        ctx.stroke();
        restore();
      }
    
      save();
      // draw the darts
      for (n=1; n<9; n++) {
      	dart(today.doy == 0 || n == today.festivalNumber);
      	ctx.rotate(deg2rad(45));
      }
      restore();
    }
    
    function doMoon(settings) {
      function moonShadow(settings) {
      	save();
      	ctx.fillStyle = grads(settings.grads)
      	ctx.rotate((day2rad(today.day) - lunaPos()));
      	ctx.beginPath();
      	ctx.arc(0, 0, settings.radius, deg2rad(-90), deg2rad(90));
      	ctx.arc(0, -settings.radius, 2, deg2rad(-90), deg2rad(90));
      	ctx.fill();
      	restore();
      }
      save();
      setStyles(settings);
      ctx.fillStyle = grads(settings.grads);
      ctx.rotate(lunaPos());
      ctx.translate(settings.radius, 0);
      ctx.beginPath();
      ctx.arc(0, 0, settings.size, 0, CIRC, true);
      if (settings.fill) ctx.fill();
      if (settings.stroke) ctx.stroke();
      moonShadow(settings.shadow);
      restore();
    }
  }
  
  //********************************** Global Helpers *******************************//
  
  // generally quicker than Math.floor()
  function floor(num) {
  	return ~~num;
  }
  
  //********************************** Settings Dialog Boxes *******************************//
  
  // create UCC Date setting dialog
  function setUCC() {
    // bail if the settings dialog is already open
    if (Id('settingsDlg')) return;
    var dlg = document.createElement('div');
    var triad = currentDate.triad;
    // ids for css targeting
    dlg.id = 'settingsDlg';
    dlg.innerHTML = ["<div id=settingsHead>Set UCC Date</div>",
                     "<div id=settingsCaptions>",
                     "<div>Day</div><div>Triad</div><div>Year</div></div>",
                     "<div id=settingsCtrls>",
                     "<div id=uccDay><input type=number min=0 max=30 id=setDay value='", currentDate.day, "' /></div>",
                     "<div id=uccTriad><select id=setTriad>",
                     "<option value='0' ", (triad == 0) ? ' selected' : '', ">Zero</option>",
                     "<option value='1' ", (triad == 1) ? ' selected' : '', ">ONE-Aries\u2648</option>",
                     "<option value='2' ", (triad == 2) ? ' selected' : '', ">TWO-Taurus\u2649</option>",
                     "<option value='3' ", (triad == 3) ? ' selected' : '', ">THREE-Gemini\u264A</option>",
                     "<option value='4' ", (triad == 4) ? ' selected' : '', ">FOUR-Cancer\u264B</option>",
                     "<option value='5' ", (triad == 5) ? ' selected' : '', ">FIVE-Leo\u264C</option>",
                     "<option value='6' ", (triad == 6) ? ' selected' : '', ">SIX-Virgo\u264D</option>",
                     "<option value='7' ", (triad == 7) ? ' selected' : '', ">SEVEN-Libra\u264E</option>",
                     "<option value='8' ", (triad == 8) ? ' selected' : '', ">EIGHT-Scorpio\u264F</option>",
                     "<option value='9' ", (triad == 9) ? ' selected' : '', ">NINE-Sagittarius\u2650</option>",
                     "<option value='10' ", (triad == 10) ? ' selected' : '', ">TEN-Capricorn\u2651</option>",
                     "<option value='11' ", (triad == 11) ? ' selected' : '', ">ELEVEN-Aquarius\u2652</option>",
                     "<option value='12' ", (triad == 12) ? ' selected' : '', ">TWELVE-Pisces\u2653</option>",
                     "</select></div>",
                     "<div id=uccYear><input type=number id=setYear value='", currentDate.year, "' /></div></div>",
                     "<div id=settingsButtons><input type=button id=btnCancel value=Cancel>",
                     "<input type=button id=btnSet value=Set></div>"].join('');
    document.body.appendChild(dlg);
    Id('btnCancel').addEventListener('click', closeSetting);
    Id('btnSet').addEventListener('click', setUCCDate);
  }
  
  // create Gregorian Date setting dialog
  function setGreg() {
    // bail if the settings dialog is already open
    if (Id('settingsDlg')) return;
    var dlg = document.createElement('div');
    var gregDate = currentDate.jDate;
    var month = gregDate.getMonth();
    // ids for css targeting
    dlg.id = 'settingsDlg';
    dlg.innerHTML = ["<div id=settingsHead>Set Gregorian Date</div>",
                     "<div id=settingsCaptions>",
                     "<div>Day</div><div>Month</div><div>Year</div></div>",
                     "<div id=settingsCtrls>",
                     "<div id=greDay><input type=number min=0 max=30 id=setDay value='", gregDate.getDate(), "' /></div>",
                     "<div id=greMonth><select id=setMonth>",
                     "<option value='0' ", (month == 0) ? ' selected' : '', ">January</option>",
                     "<option value='1' ", (month == 1) ? ' selected' : '', ">February</option>",
                     "<option value='2' ", (month == 2) ? ' selected' : '', ">March</option>",
                     "<option value='3' ", (month == 3) ? ' selected' : '', ">April</option>",
                     "<option value='4' ", (month == 4) ? ' selected' : '', ">May</option>",
                     "<option value='5' ", (month == 5) ? ' selected' : '', ">June</option>",
                     "<option value='6' ", (month == 6) ? ' selected' : '', ">July</option>",
                     "<option value='7' ", (month == 7) ? ' selected' : '', ">August</option>",
                     "<option value='8' ", (month == 8) ? ' selected' : '', ">September</option>",
                     "<option value='9' ", (month == 9) ? ' selected' : '', ">October</option>",
                     "<option value='10' ", (month == 10) ? ' selected' : '', ">November</option>",
                     "<option value='11' ", (month == 11) ? ' selected' : '', ">December</option>",
                     "</select></div>",
                     "<div id=greYear><input type=number id=setYear value='", gregDate.getFullYear(), "' /></div></div>",
                     "<div id=settingsButtons><input type=button id=btnCancel value=Cancel>",
                     "<input type=button id=btnSet value=Set></div>"].join('');
    document.body.appendChild(dlg);
    Id('btnCancel').addEventListener('click', closeSetting);
    Id('btnSet').addEventListener('click', setGregDate);
  }
  
  //******************************* Options and titles updating ****************************//
  
  // return current checkbox and option states
  function currOptions() {
  	return {
  	  language: Id('ctl_lang').value,
  	  stepDays: Id('ctl_step').value,
  	  dateFmt: Id('ctl_dateFmt').value,
  	  sandhis: Id('ctl_sandhis').checked,
  	  vrp: Id('ctl_vrp').checked,
  	  animate: Id('ctl_animate').checked,
  	  reverse: Id('ctl_reverse').checked,
  	  yearStart: Id('ctl_yearStart').checked,
  	  moon: Id('ctl_moon').checked,
  	  deekSymb: Id('ctl_deek').checked,
  	  gysf: Id('ctl_gysf').checked,
  	  constMarkers: false
  	}
  }
  
  // update the controls from saved options
  function setOptions() {
    Id('ctl_sandhis').checked = options.sandhis;
    Id('ctl_vrp').checked = options.vrp;
    Id('ctl_animate').checked = options.animate;
    Id('ctl_reverse').checked = options.reverse;
    Id('ctl_yearStart').checked = options.yearStart;
    Id('ctl_moon').checked = options.moon;
    Id('ctl_deek').checked = options.deekSymb;
    Id('ctl_gysf').checked = options.gysf;
    Id('ctl_lang').value = options.language;
    Id('ctl_dateFmt').value = options.dateFmt;
    Id('ctl_step').value = options.stepDays;
  }
  
  // update checkbox titles
  function updateTitles() {
  	var options = currOptions();
  	if (options.animate) Id('ctl_animate').title = 'Turn animation off';
  	else Id('ctl_animate').title = 'Turn animation on';
    if (options.sandhis) Id('ctl_sandhis').title = 'Turn sandhi lines off';
    else Id('ctl_sandhis').title = 'Turn sandhi lines on';
  	if (options.moon) Id('ctl_moon').title = 'Turn moon off';
  	else Id('ctl_moon').title = 'Turn moon on';
  	if (options.vrp) Id('ctl_vrp').title = 'Turn Variable Rate of Precession off';
  	else Id('ctl_vrp').title = 'Turn Variable Rate of Precession on';
  	if (options.vrp) Id('ctl_vrp').title = 'Turn Variable Rate of Precession off';
  	else Id('ctl_vrp').title = 'Turn Variable Rate of Precession on';
  	if (options.gysf) Id('ctl_gysf').title = 'Turn Great-Year-So-Far off';
  	else Id('ctl_vrp').title = 'Turn Great-Year-So-Far on';
  }
  
  //******************************************* Animation Callbacks ***************************//
  
  // calculate the actual skipSize accounting for intercalary days and leapyears
  function nextStep(reverse) {
    options = currOptions();
    reverse = reverse || options.reverse && options.animate;
    // day
    if (options.stepDays == 1)
      return (reverse ?
        new UCCDate(currentDate.year, currentDate.triad, currentDate.day - 1) : 
        new UCCDate(currentDate.year, currentDate.triad, currentDate.day + 1));
    // year
    if (options.stepDays == 365)
      return (reverse ?
        new UCCDate(currentDate.year - 1, currentDate.triad, currentDate.day) : 
        new UCCDate(currentDate.year + 1, currentDate.triad, currentDate.day));
    // triad
    if (options.stepDays == 30)
      if (reverse)
        return ( currentDate.triad > 1 ?
        new UCCDate(currentDate.year, currentDate.triad - 1, currentDate.day) : 
        new UCCDate(currentDate.year - 1, 12, currentDate.day));
      else
        return ( currentDate.triad < 12 ?
        new UCCDate(currentDate.year, currentDate.triad + 1, currentDate.day) : 
        new UCCDate(currentDate.year + 1, 1, currentDate.day));
    // decan: subtract any intercals in the intervening deek
    var d = new UCCDate(currentDate.year, currentDate.triad, (reverse ? currentDate.day - 10 : currentDate.day + 10));
    // not New Year period
    if (currentDate.doy > 12 && currentDate.doy < 355)
      // no intercals in deek
      if (d.day % 10 == currentDate.day % 10) return d;
      // deek includes an intercalary day
      else return new UCCDate(d.year, d.triad, (reverse ? d.day - 1 : d.day + 1));
    // account for new year intercals and Leap Year's Day if a leap year
    if (reverse)
      return (currentDate.day < 11 ?
        new UCCDate(currentDate.year - 1, 12, currentDate.day + 20) :
        new UCCDate(currentDate.year, 12, currentDate.day - 10));
    return (currentDate.day < 11 ?
      new UCCDate(currentDate.year, 1, currentDate.day + 10) :
      new UCCDate(currentDate.year + 1, 1, currentDate.day - 20));
  }
  
  // callback to do animation
  function animate() {
    drawCalendar(nextStep());
  }
  
  //******************************************** Touch/click Event Handling *******************************//
  
  // respond to drag start
  function dragStart(e) {
    var ctl = Id('controls');
    var box = Id('touchPanel');
    // display the controls box if hidden.
    if (ctl.style.display == '' || ctl.style.display == 'none') {
      ctl.style.display = 'block';
    }
    // add listeners for mouseup and mousemove or touchend and touchmove but not both
    if (e.type == 'mousedown') {
      box.addEventListener('mousemove', dragMove, false);
      ctl.addEventListener('mousemove', dragMove, false);
      window.addEventListener('mouseup', dragEnd, false);
    } else {
      // ignore if there is more than 1 touch
      if (e.touches.length == 1) {
        box.addEventListener('touchmove', dragMove, false);
        ctl.addEventListener('touchmove', dragMove, false);
        window.addEventListener('touchend', dragEnd, false);
      } else {
        box.removeEventListener('touchmove', dragMove, false);
        ctl.removeEventListener('touchmove', dragMove, false);
        window.removeEventListener('touchend', dragEnd, false);
        var ctlRect = ctl.getBoundingClientRect();
        ctl.style.left = -ctlRect.width + 'px';
        ctl.style.display = 'none';
      }
    }
  }
  
  // respond to drag end
  function dragEnd(e) {
    var box = Id('touchPanel');
    var ctl = Id('controls');
    var cnv = Id('clockCanvas');
    // remove the listeners
    if (e.type == 'mouseup') {
      box.removeEventListener('mousemove', dragMove, false);
      ctl.removeEventListener('mousemove', dragMove, false);
      window.removeEventListener('mouseup', dragEnd, false);
    } else {
      box.removeEventListener('touchmove', dragMove, false);
      ctl.removeEventListener('touchmove', dragMove, false);
      window.removeEventListener('touchend', dragEnd, false);
    }
    // see if the drag was far enough
    var ctlRect = ctl.getBoundingClientRect();
    var cnvRect = cnv.getBoundingClientRect();
    if (ctlRect.left > cnvRect.left - (ctlRect.width * 0.5)) {
      ctl.style.left = 10 + 'px';
    } else {
      ctl.style.left = -ctlRect.width + 'px';
      ctl.style.display = 'none';
    }
  }
  
  function dragMove(e) {
    function unZoom(x) {
      return floor(x / zoomFactor);
    }
    var ctl = Id('controls');
    var box = Id('touchPanel');
    var cnv = Id('clockCanvas');
    var ctlRect = ctl.getBoundingClientRect();
    var cnvRect = cnv.getBoundingClientRect();
    // gecko zooms mouse coords AND box sizes, webkit only mouse coords
    var ctlX = (Math.round(ctlRect.width) == ctlWidth) ?
      unZoom(e.clientX) - (cnvRect.left + ctlRect.width) :
      unZoom(e.clientX) - unZoom(ctlRect.width + cnvRect.left);
    // convert touches to mouse events if required
    if (e.type == 'touchmove') {
      // check there is still only one touch
      if (e.touches.length > 1) { return;}
      ctlX = unZoom(e.changedTouches[0].clientX) - (cnvRect.left + ctlRect.width);
    }
    // move the controls panel so it follows the pointer
    ctl.style.left = ctlX + 'px';
  }
  
  // set up touch/mouse event listeners on the touch panel and control panel
  function eventHandlers() {
    // on page load add handlers
  	window.addEventListener('load', function() {
  	  var elm = Id('touchPanel');
      elm.addEventListener('touchstart', dragStart, false);
      elm.addEventListener('mousedown', dragStart, false);
  	  elm = Id('controls');
      elm.addEventListener('touchstart', dragStart, false);
      elm.addEventListener('mousedown', dragStart, false);
    }, false);
    // on page unload save options
    window.addEventListener('unload', function() {
      localStorage.options = JSON.stringify(currOptions());
    }, false);
  }
  
  //************************************** Redraw Events ************************************//
  
  // redraw the clock, start or stop animation
  function redraw(date=currentDate) {
    options = currOptions();
    currentDate = date;  // update currentDate if a different date has been provided
    if (options.animate) {
      // switch animation on if not already running
      if (!callBack) {
        if (options.yearStart)
          if (options.reverse)
            // start at 1st day of this year
            currentDate = new UCCDate(date.year, 0, (date.leapYear ? 0 : 1));
          else
            // start at last day of previous year
            currentDate = new UCCDate(date.year - 1, 12, 30);
        callBack = window.setInterval(animate, animSpeed);
      }
    } else {
      // stop animation if it's running
      if (callBack) {
        window.clearInterval(callBack);
        callBack = 0;
      }
      drawCalendar(date);
    }
  }
  
  // rewrite redraw request events as simple function calls, to not clobber redraw() parameter
  function refresh(evt) {
    redraw();
  }
  
  // cancel button clicked, close the settings dialog
  function closeSetting(evt) {
    document.body.removeChild(Id('settingsDlg'));
  }
  
  // set button clicked, save the settings and close the dialog
  function setUCCDate(evt) {
  	currentDate = new UCCDate(Id('setYear').value, Id('setTriad').value, Id('setDay').value);
    document.body.removeChild(Id('settingsDlg'));
  	redraw();
  }
  
  // set button clicked, save the settings and close the dialog
  function setGregDate(evt) {
  	currentDate = new UCCDate(Date.UTC(Id('setYear').value, Id('setMonth').value, Id('setDay').value));
    document.body.removeChild(Id('settingsDlg'));
  	redraw();
  }
  
  // stop animation and restart in opposite direction
  function setReverse() {
    if (currOptions().animate) {
      // stop the animation if running
      if (callBack) window.clearInterval(callBack);
      // restart the animation
      callBack = window.setInterval(animate, animSpeed);
    }
  }
  
  function setToday(evt) {
  	if (currOptions().animate) return; // ignore if animating
  	currentDate = new UCCDate();
  	redraw();
  }
  
  function goBack(evt) {
  	if (currOptions().animate) return; // ignore if animating
    redraw(nextStep(true));
  }
  
  function goFwd(evt) {
  	if (currOptions().animate) return; // ignore if animating
    redraw(nextStep());
  }
  
  // publish event handling methods so they can be called by the DOM
  this.refresh = refresh;
  this.setReverse = setReverse;
  this.setUCC = setUCC;
  this.setGreg = setGreg;
  this.setToday = setToday;
  this.goBack = goBack;
  this.goFwd = goFwd;
}());